@@ -190,22 +190,6 @@ class CaptureStmt:
         self.arg_name = name
 
 
-def get_save_root(path: Path, root: Optional[Path] = None, build_root: Optional[Path] = None):
-    if root is not None:
-        relative_parts = path.parent.relative_to(root)
-        import_parts = list(relative_parts.parts)
-    else:
-        import_parts = loader.try_capture_import_parts(path)
-    if import_parts is None:
-        raise NotImplementedError("you must use inline "
-                                  "in a standard python project with "
-                                  "pip installed.")
-    if build_root is not None:
-        res = build_root / "/".join(import_parts)
-    else:
-        res = PCCM_INLINE_LIBRARY_PATH / "/".join(import_parts)
-    return res
-
 
 def get_save_file(path: Path):
     pid = os.getpid()
@@ -313,6 +297,12 @@ class InlineClass(Class):
                          "string", "iostream", "fstream")
         self.add_dependency(PyBind11)
 
+class _ModuleMeta:
+    def __init__(self, func: Any, captures: List[CaptureStmt], capture_ctypes: List[BaseType], inner_code: str) -> None:
+        self.func = func
+        self.captures = captures
+        self.capture_ctypes = capture_ctypes
+        self.inner_code = inner_code
 
 class InlineBuilder:
     """
@@ -324,8 +314,13 @@ class InlineBuilder:
             plugins: Optional[Dict[str, InlineBuilderPlugin]] = None,
             root: Optional[Path] = None,
             build_root: Optional[Path] = None,
-            build_kwargs: Optional[Dict[str, Any]] = None) -> None:
+            build_kwargs: Optional[Dict[str, Any]] = None,
+            param_deps: Optional[List[pccm.ParameterizedClass]] = None,
+            reload_when_code_change: bool = False) -> None:
         self.deps = deps
+        if param_deps is None:
+            param_deps = []
+        self.param_deps = param_deps
         if plugins is None:
             self.plugins = _DEFAULT_PLUGINS
         else:
@@ -333,11 +328,11 @@ class InlineBuilder:
         if build_kwargs is None:
             build_kwargs = {}
         self.build_kwargs = build_kwargs
-        self.module_functions: Dict[Tuple[str, str, int], Any] = {}
-        self.cached_captures: Dict[Tuple[str, str, int],
-                                   List[CaptureStmt]] = {}
-        self.cached_capture_ctypes: Dict[Tuple[str, str, int],
-                                         List[BaseType]] = {}
+        self.modules: Dict[Tuple[str, str, int], _ModuleMeta] = {}
+        # self.cached_captures: Dict[Tuple[str, str, int],
+        #                            List[CaptureStmt]] = {}
+        # self.cached_capture_ctypes: Dict[Tuple[str, str, int],
+        #                                  List[BaseType]] = {}
 
         self.lock = threading.Lock()
 
@@ -346,6 +341,7 @@ class InlineBuilder:
         self.used_names: Set[Tuple[str, str]] = set()
         self.root = root
         self.build_root = build_root
+        self._reload_when_code_change = reload_when_code_change
 
     def _find_exist_module_name(self, code: str, code_hash: str, root: Path):
         cur_dep_ids = self.dep_ids
@@ -357,6 +353,23 @@ class InlineBuilder:
                 return path.stem
         return None
 
+    def get_save_root(self, path: Path, root: Optional[Path] = None, build_root: Optional[Path] = None):
+        if root is not None:
+            relative_parts = path.resolve().parent.relative_to(root)
+            import_parts = list(relative_parts.parts)
+        else:
+            import_parts = loader.try_capture_import_parts(path)
+        if import_parts is None:
+            raise NotImplementedError("you must use inline "
+                                    "in a standard python project with "
+                                    "pip installed.")
+        if build_root is not None:
+            res = build_root / "/".join(import_parts)
+        else:
+            res = PCCM_INLINE_LIBRARY_PATH / "/".join(import_parts)
+        return res
+
+
     def handle_container_code(self, code_str: str, code: FunctionCode,
                     arg: Optional[Any]):
         code.raw(code_str)
@@ -366,6 +379,29 @@ class InlineBuilder:
         
         return None 
 
+    def build(self, pccm_cls: pccm.Class, mod_root: Path, name: str, timeout: float, user_arg: Optional[Any] = None):
+        mod_root.mkdir(mode=0o755, parents=True, exist_ok=True)
+        out_lib_path = mod_root / name
+        build_dir = mod_root / name
+        # out_lib_meta_path = mod_root / f"{prev_mod_name}.json"
+        file_lock = mod_root / f"{name}.lock"
+        # -1 is invalid for portalocker
+        with portalocker.Lock(str(file_lock), timeout=timeout) as fh:
+            mod = build_pybind([pccm_cls],
+                                out_lib_path,
+                                build_dir=build_dir,
+                                **self.build_kwargs)
+
+        return getattr(
+            getattr(getattr(mod, PCCM_INLINE_NAMESPACE),
+                    PCCM_INLINE_CLASS_NAME), PCCM_INLINE_FUNCTION_NAME)
+
+    def run_func(self, func, *args, user_args: Optional[Any] = None):
+        return func(*args)
+
+    def get_base_class(self):
+        return InlineClass()
+
     def inline(self,
                name: str,
                code: Union[str, FunctionCode],
@@ -374,8 +410,10 @@ class InlineBuilder:
                *,
                _frame_cnt: int = 1,
                user_arg: Optional[Any] = None,
-               timeout: float=999999.0):
+               timeout: float=999999.0,
+               disable_cache: bool = False):
         """use $var to capture python objects, use $(var.shape[0]) to capture anonymous expr.
+        use different to handle different arg types.
         ~20-100us run overhead. 
         only support: 
         1. int/float/str and nested containers of int/float/str.
@@ -395,22 +433,24 @@ class InlineBuilder:
             assert frame is not None
             cur_frame = frame
             _frame_cnt -= 1
-        del frame
+        # del frame
         local_vars = cur_frame.f_locals.copy()
         local_vars.update(additional_vars)
         code_path = cur_frame.f_code.co_filename
-        lineno = cur_frame.f_code.co_firstlineno
+        lineno = cur_frame.f_lineno
         key = (code_path, name)
         unique_key = (code_path, name, lineno)
-
-        del cur_frame
+        if self._reload_when_code_change:
+            unique_key = (code_path, name, -1)
+        # del cur_frame
         with self.lock:
-            exist = unique_key in self.module_functions
-            if key in self.used_names and not exist:
-                raise ValueError("you use duplicate name in same file.")
+            if not disable_cache and not self._reload_when_code_change:
+                exist = unique_key in self.modules
+                if key in self.used_names and not exist:
+                    raise ValueError("you use duplicate name in same file.", unique_key)
+            else:
+                exist = False
             if not exist:
-                mod_root = get_save_root(Path(code_path), self.root, self.build_root)
-                mod_root.mkdir(mode=0o755, parents=True, exist_ok=True)
                 # 1. extract captured vars
                 it = source_iter.CppSourceIterator(code_str)
                 # hold ranges for further replace
@@ -467,8 +507,12 @@ class InlineBuilder:
                         # eval expr in prev frame
                         obj = eval(cap.name, local_vars)
                     # apply non-anonymous vars (expr are anonymous vars)
-                    cpp_type, mapped_cpp_type = nested_type_analysis(
-                        obj, self.plugins, user_arg=user_arg)
+                    try:
+                        cpp_type, mapped_cpp_type = nested_type_analysis(
+                            obj, self.plugins, user_arg=user_arg)
+                    except:
+                        print(f"ERROR: variable {cap.name} type analysis failed.")
+                        raise
                     obj = _nested_apply_plugin_transform(
                         cpp_type, obj, self.plugins, user_arg)
                     if cap.is_expr:
@@ -503,82 +547,100 @@ class InlineBuilder:
                     args.append(v)
 
                 inner_code_str = execute_modifiers(code_str, replaces)
+
                 assert inner_code_str is not None
+                if self._reload_when_code_change:
+                    exist = unique_key in self.modules
+                    if exist:
+                        prev_inner_code = self.modules[unique_key].inner_code
+                        if prev_inner_code == inner_code_str:
+                            module_meta = self.modules[unique_key]
+                            func = module_meta.func
+                            all_captures = module_meta.captures
+                            all_base_types = module_meta.capture_ctypes
+                            args = []
+                            for cap, bt in zip(all_captures, all_base_types):
+                                if not cap.is_expr:
+                                    if cap.name not in local_vars:
+                                        raise ValueError(
+                                            f"can't find your capture {cap.name} in prev frame.")
+                                    obj = local_vars[cap.name]
+                                else:
+                                    for cap_name in cap.expr_names:
+                                        if cap_name not in local_vars:
+                                            raise ValueError(
+                                                f"can't find your capture {cap_name} in prev frame."
+                                            )
+                                    # eval expr in prev frame
+                                    obj = eval(cap.name, local_vars)
+                                obj = _nested_apply_plugin_transform(bt, obj, self.plugins, user_arg)
+                                args.append(obj)
+                            for v in additional_vars.values():
+                                args.append(v)
+                            return self.run_func(func, *args, user_args=user_arg)
                 if isinstance(code, FunctionCode):
                     container_fcode.ret(code.return_type)
-                self.handle_container_code(inner_code_str, container_fcode, user_arg)
+                meta = self.handle_container_code(inner_code_str, container_fcode, user_arg)
                 inner_decl = self.create_inner_decl(inner_code_str, container_fcode, inner_fcode, user_arg)
 
                 # now we have complete code. we need to determine a history build dir and use it to build library if need.
                 # here we must reserve build dir because we need to rebuild when dependency change.
-                meta = StaticMemberFunctionMeta(impl_file_suffix=impl_file_suffix)
-                meta.mw_metas.append(Pybind11MethodMeta())
+                if meta is None:
+                    meta = StaticMemberFunctionMeta(impl_file_suffix=impl_file_suffix)
+                    meta.mw_metas.append(Pybind11MethodMeta())
                 decl = FunctionDecl(meta, container_fcode)
                 decl.meta.name = PCCM_INLINE_FUNCTION_NAME
                 # container_fcode_str = decl.inspect_impl()
                 # container_fcode_hash = hashlib.sha256(code_str.encode('utf-8')).hexdigest()
 
-                pccm_class = InlineClass()
+                pccm_class = self.get_base_class()
                 pccm_class.class_name = PCCM_INLINE_CLASS_NAME
                 pccm_class.add_func_decl(decl)
 
                 for dep in decl.code._impl_only_deps:
                     pccm_class.add_impl_only_dependency_by_name(
-                        decl.meta.name, dep)
+                        decl.meta.name, dep.get_class_type())
                 if isinstance(code, FunctionCode):
                     for dep in code._impl_only_deps:
                         pccm_class.add_impl_only_dependency_by_name(
-                            decl.meta.name, dep)
+                            decl.meta.name, dep.get_class_type())
 
                 if inner_decl is not None:
                     inner_decl.meta.name = PCCM_INLINE_INNER_FUNCTION_NAME
                     pccm_class.add_func_decl(inner_decl)
                     for dep in inner_decl.code._impl_only_deps:
                         pccm_class.add_impl_only_dependency_by_name(
-                            inner_decl.meta.name, dep)
+                            inner_decl.meta.name, dep.get_class_type())
                     if isinstance(code, FunctionCode):
                         for dep in code._impl_only_deps:
                             pccm_class.add_impl_only_dependency_by_name(
-                                inner_decl.meta.name, dep)
+                                inner_decl.meta.name, dep.get_class_type())
 
                 pccm_class.add_dependency(*self.deps)
+                for pdep in self.param_deps:
+                    pdep_ns = pdep.namespace
+                    pdep_cls_name = pdep.get_user_provided_class_name()
+                    assert pdep_ns is not None, "you must provide a namespace for param dep"
+                    assert pdep_cls_name is not None, "you must provide a class_name for param dep"
+                    # use user-defined class name as alias
+                    pccm_class.add_param_class(pdep_cls_name, pdep, pdep_cls_name)
                 pccm_class.namespace = PCCM_INLINE_NAMESPACE
-                # name format: code_hash + line + pid
-                # if name == "":
-                #     prev_mod_name = self._find_exist_module_name(container_fcode_str, container_fcode_hash, mod_root)
-                # else:
-                # to avoid multi-process problem, we need to
-                prev_mod_name = name
-                # if prev_mod_name is None:
-                #     prev_mod_name = f"_{container_fcode_hash}_{lineno}_{os.getpid()}"
-                #     out_lib_meta_path = mod_root / f"{prev_mod_name}.json"
-                #     meta = ModuleMetaData(container_fcode_str, self.dep_ids)
-                #     with out_lib_meta_path.open("w") as f:
-                #         json.dump(dataclasses.asdict(meta), f)
-                out_lib_path = mod_root / prev_mod_name
-                build_dir = mod_root / prev_mod_name
-                # out_lib_meta_path = mod_root / f"{prev_mod_name}.json"
-                file_lock = mod_root / f"{prev_mod_name}.lock"
-                # -1 is invalid for portalocker
-                with portalocker.Lock(str(file_lock), timeout=timeout) as fh:
-                    mod = build_pybind([pccm_class],
-                                       out_lib_path,
-                                       build_dir=build_dir,
-                                       **self.build_kwargs)
-
-                self.module_functions[unique_key] = getattr(
-                    getattr(getattr(mod, PCCM_INLINE_NAMESPACE),
-                            PCCM_INLINE_CLASS_NAME), PCCM_INLINE_FUNCTION_NAME)
-                self.cached_captures[unique_key] = all_captures
-                self.cached_capture_ctypes[unique_key] = capture_bts
-                self.used_names.add(key)
-                # breakpoint()
-                return self.module_functions[unique_key](*args)
+                mod_root = self.get_save_root(Path(code_path), self.root, self.build_root)
+                func = self.build(pccm_class, mod_root, name, timeout, user_arg)
+                if not disable_cache:
+                    module_meta = _ModuleMeta(func, all_captures, capture_bts, inner_code_str)
+                    self.modules[unique_key] = module_meta
+                    # self.cached_captures[unique_key] = all_captures
+                    # self.cached_capture_ctypes[unique_key] = capture_bts
+                    self.used_names.add(key)
+                    # breakpoint()
+                return self.run_func(func, *args, user_args=user_arg)
 
         # module already loaded. just run it after transform.
-        func = self.module_functions[unique_key]
-        all_captures = self.cached_captures[unique_key]
-        all_base_types = self.cached_capture_ctypes[unique_key]
+        module_meta = self.modules[unique_key]
+        func = module_meta.func
+        all_captures = module_meta.captures
+        all_base_types = module_meta.capture_ctypes
         args = []
         for cap, bt in zip(all_captures, all_base_types):
             if not cap.is_expr:
@@ -598,7 +660,7 @@ class InlineBuilder:
             args.append(obj)
         for v in additional_vars.values():
             args.append(v)
-        return func(*args)
+        return self.run_func(func, *args, user_args=user_arg)
 
 
 def main():
