@@ -9,8 +9,8 @@ from pccm.core import (Class, CodeSectionClassDef, ConstructorMeta, EnumClass,
                        ExternalFunctionMeta, FunctionCode, FunctionDecl,
                        ManualClass, ManualClassGenerator, Member,
                        MemberFunctionMeta, MiddlewareMeta, ParameterizedClass,
-                       StaticMemberFunctionMeta, get_class_meta)
-from pccm.core.buildmeta import _unique_list_keep_order
+                       StaticMemberFunctionMeta, UnifiedDependency, get_class_meta)
+from ccimport.buildmeta import _unique_list_keep_order
 from pccm.core.codegen import Block, generate_code, generate_code_list
 from pccm.core.markers import middleware_decorator, append_class_mw_meta
 
@@ -173,6 +173,9 @@ class TemplateTypeStmt(object):
         "tv::array":
         lambda args: "List[{}]".format(args[0].to_pyanno()),
         "std::tuple":
+        lambda args: "Tuple[{}]".format(", ".join(a.to_pyanno()
+                                                  for a in args)),
+        "std::pair":
         lambda args: "Tuple[{}]".format(", ".join(a.to_pyanno()
                                                   for a in args)),
         "std::vector":
@@ -189,6 +192,10 @@ class TemplateTypeStmt(object):
         lambda args: "Set[{}]".format(args[0].to_pyanno()),
         "std::unordered_set":
         lambda args: "Set[{}]".format(args[0].to_pyanno()),
+        "std::shared_ptr":
+        lambda args: args[0].to_pyanno(),
+        "std::unique_ptr":
+        lambda args: args[0].to_pyanno(),
     }  # type: Dict[str, Callable[[List["TemplateTypeStmt"]], str]]
 
     def __init__(self,
@@ -231,13 +238,21 @@ def _simple_template_type_parser_recursive(
     is_ptr = False 
     if stmt.endswith("__restrict__"):
         end -= _LEN_RESTRICT
-        stmt = stmt[:end].strip()
+        stmt = stmt[:end]
+        stmt2 = stmt.strip()
+        end -= len(stmt) - len(stmt2)
+        stmt = stmt2
     if stmt[-1] == "*":
         is_ptr = True 
         end -= 1
     elif stmt[-1] == "&":
         end -= 1
-    stmt = stmt[:end].strip()
+
+    stmt = stmt[:end]
+    stmt2 = stmt.strip()
+    end -= len(stmt) - len(stmt2)
+    stmt = stmt2
+
     if stmt[end - 1] != ">":
         # type with no template param
         name = stmt[begin:end].strip()
@@ -324,8 +339,9 @@ class Pybind11Meta(MiddlewareMeta):
     pass
 
 class Pybind11ClassMeta:
-    def __init__(self, module_local: bool = False):
+    def __init__(self, module_local: bool = False, is_shared_class: bool = False):
         self.module_local = module_local
+        self.is_shared_class = is_shared_class
 
 class ReturnPolicy(Enum):
     TakeOwnerShip = "pybind11::return_value_policy::take_ownership"
@@ -379,7 +395,7 @@ class Pybind11PropMeta(Pybind11Meta):
 class Pybind11ClassMwMeta(Pybind11Meta):
     """may add some attributes in future.
     """
-    def __init__(self, raw_code: FunctionCode):
+    def __init__(self, raw_code: Optional[FunctionCode] = None):
         super().__init__(Pybind11SplitImpl)
         self.raw_code = raw_code
 
@@ -574,7 +590,8 @@ def _postprocess_class(cls_name: str,
                        enum_classes: List[EnumClass],
                        parent_ns: str = "",
                        module_local: bool = False,
-                       raw_defs: Optional[List[FunctionCode]] = None):
+                       raw_defs: Optional[List[FunctionCode]] = None,
+                       is_shared_class: bool = False):
     has_virtual = False
     if raw_defs is None:
         raw_defs = []
@@ -638,6 +655,8 @@ def _postprocess_class(cls_name: str,
             break
     cls_qual_name = "{}::{}".format(cls_namespace.replace(".", "::"), cls_name)
     cls_def_arguments = [cls_qual_name]
+    if is_shared_class:
+        cls_def_arguments.append(f"std::shared_ptr<{cls_qual_name}>")
     if has_virtual:
         cls_def_arguments.append("{}::{}".format(
             cls_namespace.replace(".", "::"), virtual_cls_name))
@@ -783,9 +802,7 @@ def _generate_python_interface_class(cls_name: str,
         prop decls
         methods (overloaded methods)
     TODO handle c++ operators
-    TODO better code
-    TODO auto generate STL annotations
-    TODO insert docstring if exists
+    TODO better codegenerat
     """
     imports = []  # type: List[str]
     # split decl to method decl and prop decl
@@ -910,7 +927,9 @@ def _generate_python_interface_class(cls_name: str,
     class_def_string = "class {}:".format(cls_name)
     if parent_class_name:
         class_def_string = "class {}({}):".format(cls_name, parent_class_name)
-        imports.append("from {} import {}".format(parent_import_ns, parent_class_name))
+        if parent_import_ns:
+            imports.append("from {} import {}".format(parent_import_ns, parent_class_name))
+    
     class_block = Block(class_def_string, decl_codes)
     return class_block, imports
 
@@ -947,19 +966,21 @@ class Pybind11SingleClassHandler(ManualClass):
         assert cu.namespace is not None
         if mw_meta.raw_bind_anno:
             self.raw_bind_annos.append(mw_meta.raw_bind_anno)
-        self.func_decls.append(
-            PybindMethodDecl(func_decl, cu.namespace, cu.class_name, mw_meta))
+        decl = PybindMethodDecl(func_decl, cu.namespace, cu.class_name, mw_meta)
+        if not mw_meta.prop_name:
+            mw_meta.prop_name = decl.func_name
+        self.func_decls.append(decl)
 
     def handle_member(self, cu: Class, member_decl: Member,
                       mw_meta: Pybind11PropMeta):
         assert cu.namespace is not None
-
         self.prop_decls.append(
             PybindPropDecl(member_decl, cu.namespace, cu.class_name, mw_meta))
 
     def handle_class_meta(self, cu: Class, mw_meta: Pybind11ClassMwMeta):
         assert cu.namespace is not None
-        self.raw_defs.append(mw_meta.raw_code)
+        if mw_meta.raw_code is not None:
+            self.raw_defs.append(mw_meta.raw_code)
 
     def postprocess(self, parent_is_pybind: bool = False):
         if self.built:
@@ -968,20 +989,23 @@ class Pybind11SingleClassHandler(ManualClass):
         bind_code.arg("module", "const pybind11::module_&")
         func_meta = StaticMemberFunctionMeta(impl_file_suffix=self.file_suffix)
         parent_name = "" if not parent_is_pybind else self.cu.get_parent_name()
-        cu_type = type( self.cu)
+        cu_type = type(self.cu)
         cls_meta=  get_class_meta(cu_type)
         module_local = False 
+        shared_ptr_class = False
         if cls_meta is not None:
             for mw_meta in cls_meta.mw_metas:
                 if isinstance(mw_meta, Pybind11ClassMeta):
                     module_local = mw_meta.module_local
+                    shared_ptr_class = mw_meta.is_shared_class
         cls_def_block, vblock = _postprocess_class(self.cu.class_name,
                                                    self.cu.namespace, "module",
                                                    self.get_pybind_decls(),
                                                    self.cu._enum_classes,
                                                    parent_name,
                                                    module_local,
-                                                   self.raw_defs)
+                                                   self.raw_defs,
+                                                   shared_ptr_class)
         func_meta.name = self.bind_func_name
         if vblock is not None:
             bind_code.code_after_include = "\n".join(
@@ -1014,6 +1038,7 @@ class Pybind11SplitMain(ParameterizedClass):
         self.add_include("pybind11/stl.h")
         self.add_include("pybind11/pybind11.h")
         self.add_include("pybind11/numpy.h")
+        self.add_include("memory")
         self.file_suffix = file_suffix
         self.built = False
 
@@ -1100,6 +1125,8 @@ class Pybind11SplitMain(ParameterizedClass):
                 parent_ns = ".".join(parent_name_splits[:-1])
                 parent_import_ns = "." * num_level + parent_ns
                 parent_class_name = parent_name_splits[-1]
+                if origin_cu.namespace == parent_ns:
+                    parent_import_ns = ""
             class_block, cls_imports = _generate_python_interface_class(
                 origin_cu.class_name, bind_cu.get_pybind_decls(),
                 origin_cu._enum_classes, exist_annos, parent_import_ns, parent_class_name)
@@ -1149,7 +1176,7 @@ class Pybind11SplitImpl(ManualClassGenerator):
         bind_cu.class_name = "PyBind" + cu.class_name
         bind_cu.namespace = cu.namespace
         self.bind_cus.append(bind_cu)
-        self.main_cu._unified_deps.append(bind_cu)
+        self.main_cu._unified_deps.append(UnifiedDependency(bind_cu, buildmeta_inherit=True))
         # self.main_cu.add_param_class("bind_{}".format(bind_cu.class_name), bind_cu)
         return bind_cu
 
@@ -1222,6 +1249,10 @@ def bind_class_module_local(cls=None):
     meta = Pybind11ClassMeta(True)
     return append_class_mw_meta(cls, [meta])
 
+def bind_class_options(cls=None, module_local: bool = False, shared_ptr: bool = False):
+    meta = Pybind11ClassMeta(module_local, shared_ptr)
+    return append_class_mw_meta(cls, [meta])
+
 if __name__ == "__main__":
     # print(ast.parse)
     print(python_anno_parser("Dict[int, List[int]]"))
